@@ -2,8 +2,8 @@
 #include <Python.h>
 #include "structmember.h"
 
+#define NPY_NO_DEPRECATED_API NPY_API_VERSION
 #define _MULTIARRAYMODULE
-#define NPY_NO_PREFIX
 #include "numpy/arrayobject.h"
 #include "numpy/arrayscalars.h"
 
@@ -11,17 +11,21 @@
 
 #include "npy_config.h"
 
+#include "npy_pycompat.h"
+
 #include "ctors.h"
 #include "descriptor.h"
 #include "scalartypes.h"
 
+#include "common.h"
+
 static PyArray_Descr *
 _descr_from_subtype(PyObject *type)
 {
     PyObject *mro;
     mro = ((PyTypeObject *)type)->tp_mro;
     if (PyTuple_GET_SIZE(mro) < 2) {
-        return PyArray_DescrFromType(PyArray_OBJECT);
+        return PyArray_DescrFromType(NPY_OBJECT);
     }
     return PyArray_DescrFromTypeObject(PyTuple_GET_ITEM(mro, 1));
 }
@@ -31,7 +35,7 @@ scalar_value(PyObject *scalar, PyArray_Descr *descr)
 {
     int type_num;
     int align;
-    intp memloc;
+    npy_intp memloc;
     if (descr == NULL) {
         descr = PyArray_DescrFromScalar(scalar);
         type_num = descr->type_num;
@@ -53,6 +57,7 @@ scalar_value(PyObject *scalar, PyArray_Descr *descr)
         CASE(ULONG, ULong);
         CASE(LONGLONG, LongLong);
         CASE(ULONGLONG, ULongLong);
+        CASE(HALF, Half);
         CASE(FLOAT, Float);
         CASE(DOUBLE, Double);
         CASE(LONGDOUBLE, LongDouble);
@@ -60,6 +65,8 @@ scalar_value(PyObject *scalar, PyArray_Descr *descr)
         CASE(CDOUBLE, CDouble);
         CASE(CLONGDOUBLE, CLongDouble);
         CASE(OBJECT, Object);
+        CASE(DATETIME, Datetime);
+        CASE(TIMEDELTA, Timedelta);
 #undef CASE
         case NPY_STRING:
             return (void *)PyString_AS_STRING(scalar);
@@ -87,6 +94,7 @@ scalar_value(PyObject *scalar, PyArray_Descr *descr)
                 _IFCASE(Int);
                 _IFCASE(Long);
                 _IFCASE(LongLong);
+                _IFCASE(Timedelta);
             }
             else {
                 /* Unsigned Integer */
@@ -100,6 +108,7 @@ scalar_value(PyObject *scalar, PyArray_Descr *descr)
         else {
             /* Inexact */
             if _CHK(Floating) {
+                _IFCASE(Half);
                 _IFCASE(Float);
                 _IFCASE(Double);
                 _IFCASE(LongDouble);
@@ -115,6 +124,9 @@ scalar_value(PyObject *scalar, PyArray_Descr *descr)
     else if (_CHK(Bool)) {
         return _OBJ(Bool);
     }
+    else if (_CHK(Datetime)) {
+        return _OBJ(Datetime);
+    }
     else if (_CHK(Flexible)) {
         if (_CHK(String)) {
             return (void *)PyString_AS_STRING(scalar);
@@ -135,7 +147,7 @@ scalar_value(PyObject *scalar, PyArray_Descr *descr)
      * Use the alignment flag to figure out where the data begins
      * after a PyObject_HEAD
      */
-    memloc = (intp)scalar;
+    memloc = (npy_intp)scalar;
     memloc += sizeof(PyObject);
     /* now round-up to the nearest alignment value */
     align = descr->alignment;
@@ -148,6 +160,15 @@ scalar_value(PyObject *scalar, PyArray_Descr *descr)
 #undef _CHK
 }
 
+/*NUMPY_API
+ * return true an object is exactly a numpy scalar
+ */
+NPY_NO_EXPORT int
+PyArray_CheckAnyScalarExact(PyObject * obj)
+{
+    return is_anyscalar_exact(obj);
+}
+
 /*NUMPY_API
  * Convert to c-type
  *
@@ -209,12 +230,12 @@ PyArray_CastScalarToCtype(PyObject *scalar, void *ctypeptr,
                     outcode,
                     0, NULL,
                     NULL, ctypeptr,
-                    CARRAY, NULL);
+                    NPY_ARRAY_CARRAY, NULL);
         if (aout == NULL) {
             Py_DECREF(ain);
             return -1;
         }
-        castfunc(ain->data, aout->data, 1, ain, aout);
+        castfunc(PyArray_DATA(ain), PyArray_DATA(aout), 1, ain, aout);
         Py_DECREF(ain);
         Py_DECREF(aout);
     }
@@ -257,37 +278,40 @@ NPY_NO_EXPORT PyObject *
 PyArray_FromScalar(PyObject *scalar, PyArray_Descr *outcode)
 {
     PyArray_Descr *typecode;
-    PyObject *r;
+    PyArrayObject *r;
     char *memptr;
     PyObject *ret;
 
     /* convert to 0-dim array of scalar typecode */
     typecode = PyArray_DescrFromScalar(scalar);
-    if ((typecode->type_num == PyArray_VOID) &&
-            !(((PyVoidScalarObject *)scalar)->flags & OWNDATA) &&
+    if (typecode == NULL) {
+        return NULL;
+    }
+    if ((typecode->type_num == NPY_VOID) &&
+            !(((PyVoidScalarObject *)scalar)->flags & NPY_ARRAY_OWNDATA) &&
             outcode == NULL) {
-        r = PyArray_NewFromDescr(&PyArray_Type,
-                typecode,
+        return PyArray_NewFromDescrAndBase(
+                &PyArray_Type, typecode,
                 0, NULL, NULL,
                 ((PyVoidScalarObject *)scalar)->obval,
                 ((PyVoidScalarObject *)scalar)->flags,
-                NULL);
-        PyArray_BASE(r) = (PyObject *)scalar;
-        Py_INCREF(scalar);
-        return r;
+                NULL, (PyObject *)scalar);
     }
 
-    r = PyArray_NewFromDescr(&PyArray_Type,
+    /* Need to INCREF typecode because PyArray_NewFromDescr steals a
+     * reference below and we still need to access typecode afterwards. */
+    Py_INCREF(typecode);
+    r = (PyArrayObject *)PyArray_NewFromDescr(&PyArray_Type,
             typecode,
             0, NULL,
             NULL, NULL, 0, NULL);
     if (r==NULL) {
-        Py_XDECREF(outcode);
+        Py_DECREF(typecode); Py_XDECREF(outcode);
         return NULL;
     }
     if (PyDataType_FLAGCHK(typecode, NPY_USE_SETITEM)) {
         if (typecode->f->setitem(scalar, PyArray_DATA(r), r) < 0) {
-            Py_XDECREF(outcode); Py_DECREF(r);
+            Py_DECREF(typecode); Py_XDECREF(outcode); Py_DECREF(r);
             return NULL;
         }
         goto finish;
@@ -296,9 +320,9 @@ PyArray_FromScalar(PyObject *scalar, PyArray_Descr *outcode)
     memptr = scalar_value(scalar, typecode);
 
 #ifndef Py_UNICODE_WIDE
-    if (typecode->type_num == PyArray_UNICODE) {
+    if (typecode->type_num == NPY_UNICODE) {
         PyUCS2Buffer_AsUCS4((Py_UNICODE *)memptr,
-                (PyArray_UCS4 *)PyArray_DATA(r),
+                (npy_ucs4 *)PyArray_DATA(r),
                 PyUnicode_GET_SIZE(scalar),
                 PyArray_ITEMSIZE(r) >> 2);
     }
@@ -307,24 +331,27 @@ PyArray_FromScalar(PyObject *scalar, PyArray_Descr *outcode)
     {
         memcpy(PyArray_DATA(r), memptr, PyArray_ITEMSIZE(r));
         if (PyDataType_FLAGCHK(typecode, NPY_ITEM_HASOBJECT)) {
-	    /* Need to INCREF just the PyObject portion */
+            /* Need to INCREF just the PyObject portion */
             PyArray_Item_INCREF(memptr, typecode);
         }
     }
 
 finish:
     if (outcode == NULL) {
-        return r;
+        Py_DECREF(typecode);
+        return (PyObject *)r;
     }
-    if (outcode->type_num == typecode->type_num) {
-        if (!PyTypeNum_ISEXTENDED(typecode->type_num) ||
-                (outcode->elsize == typecode->elsize))
-            return r;
+    if (PyArray_EquivTypes(outcode, typecode)) {
+        if (!PyTypeNum_ISEXTENDED(typecode->type_num)
+                || (outcode->elsize == typecode->elsize)) {
+            Py_DECREF(typecode); Py_DECREF(outcode);
+            return (PyObject *)r;
+        }
     }
 
     /* cast if necessary to desired output typecode */
     ret = PyArray_CastToType((PyArrayObject *)r, outcode, 0);
-    Py_DECREF(r);
+    Py_DECREF(typecode); Py_DECREF(r);
     return ret;
 }
 
@@ -339,9 +366,22 @@ PyArray_ScalarFromObject(PyObject *object)
 {
     PyObject *ret=NULL;
     if (PyArray_IsZeroDim(object)) {
-        return PyArray_ToScalar(PyArray_DATA(object), object);
+        return PyArray_ToScalar(PyArray_DATA((PyArrayObject *)object),
+                                (PyArrayObject *)object);
     }
-    if (PyInt_Check(object)) {
+    /*
+     * Booleans in Python are implemented as a subclass of integers,
+     * so PyBool_Check must be called before PyInt_Check.
+     */
+    if (PyBool_Check(object)) {
+        if (object == Py_True) {
+            PyArrayScalar_RETURN_TRUE;
+        }
+        else {
+            PyArrayScalar_RETURN_FALSE;
+        }
+    }
+    else if (PyInt_Check(object)) {
         ret = PyArrayScalar_New(Long);
         if (ret == NULL) {
             return NULL;
@@ -360,15 +400,13 @@ PyArray_ScalarFromObject(PyObject *object)
         if (ret == NULL) {
             return NULL;
         }
-        PyArrayScalar_VAL(ret, CDouble).real =
-                ((PyComplexObject *)object)->cval.real;
-        PyArrayScalar_VAL(ret, CDouble).imag =
-                ((PyComplexObject *)object)->cval.imag;
+        PyArrayScalar_VAL(ret, CDouble).real = PyComplex_RealAsDouble(object);
+        PyArrayScalar_VAL(ret, CDouble).imag = PyComplex_ImagAsDouble(object);
     }
     else if (PyLong_Check(object)) {
-        longlong val;
+        npy_longlong val;
         val = PyLong_AsLongLong(object);
-        if (val==-1 && PyErr_Occurred()) {
+        if (error_converting(val)) {
             PyErr_Clear();
             return NULL;
         }
@@ -378,14 +416,6 @@ PyArray_ScalarFromObject(PyObject *object)
         }
         PyArrayScalar_VAL(ret, LongLong) = val;
     }
-    else if (PyBool_Check(object)) {
-        if (object == Py_True) {
-            PyArrayScalar_RETURN_TRUE;
-        }
-        else {
-            PyArrayScalar_RETURN_FALSE;
-        }
-    }
     return ret;
 }
 
@@ -396,11 +426,11 @@ NPY_NO_EXPORT PyArray_Descr *
 PyArray_DescrFromTypeObject(PyObject *type)
 {
     int typenum;
-    PyArray_Descr *new, *conv=NULL;
+    PyArray_Descr *new, *conv = NULL;
 
     /* if it's a builtin type, then use the typenumber */
     typenum = _typenum_fromtypeobj(type,1);
-    if (typenum != PyArray_NOTYPE) {
+    if (typenum != NPY_NOTYPE) {
         new = PyArray_DescrFromType(typenum);
         return new;
     }
@@ -409,27 +439,27 @@ PyArray_DescrFromTypeObject(PyObject *type)
     if ((type == (PyObject *) &PyNumberArrType_Type) ||
             (type == (PyObject *) &PyInexactArrType_Type) ||
             (type == (PyObject *) &PyFloatingArrType_Type)) {
-        typenum = PyArray_DOUBLE;
+        typenum = NPY_DOUBLE;
     }
     else if (type == (PyObject *)&PyComplexFloatingArrType_Type) {
-        typenum = PyArray_CDOUBLE;
+        typenum = NPY_CDOUBLE;
     }
     else if ((type == (PyObject *)&PyIntegerArrType_Type) ||
             (type == (PyObject *)&PySignedIntegerArrType_Type)) {
-        typenum = PyArray_LONG;
+        typenum = NPY_LONG;
     }
     else if (type == (PyObject *) &PyUnsignedIntegerArrType_Type) {
-        typenum = PyArray_ULONG;
+        typenum = NPY_ULONG;
     }
     else if (type == (PyObject *) &PyCharacterArrType_Type) {
-        typenum = PyArray_STRING;
+        typenum = NPY_STRING;
     }
     else if ((type == (PyObject *) &PyGenericArrType_Type) ||
             (type == (PyObject *) &PyFlexibleArrType_Type)) {
-        typenum = PyArray_VOID;
+        typenum = NPY_VOID;
     }
 
-    if (typenum != PyArray_NOTYPE) {
+    if (typenum != NPY_NOTYPE) {
         return PyArray_DescrFromType(typenum);
     }
 
@@ -440,7 +470,7 @@ PyArray_DescrFromTypeObject(PyObject *type)
 
     /* Do special thing for VOID sub-types */
     if (PyType_IsSubtype((PyTypeObject *)type, &PyVoidArrType_Type)) {
-        new = PyArray_DescrNewFromType(PyArray_VOID);
+        new = PyArray_DescrNewFromType(NPY_VOID);
         conv = _arraydescr_fromobj(type);
         if (conv) {
             new->fields = conv->fields;
@@ -479,7 +509,7 @@ PyArray_FieldNames(PyObject *fields)
     if (_numpy_internal == NULL) {
         return NULL;
     }
-    tup = PyObject_CallMethod(_numpy_internal, "_makenames_list", "O", fields);
+    tup = PyObject_CallMethod(_numpy_internal, "_makenames_list", "OO", fields, Py_False);
     Py_DECREF(_numpy_internal);
     if (tup == NULL) {
         return NULL;
@@ -506,31 +536,50 @@ PyArray_DescrFromScalar(PyObject *sc)
         Py_INCREF(descr);
         return descr;
     }
-    
-    descr = PyArray_DescrFromTypeObject((PyObject *)sc->ob_type);
-    if (descr->elsize == 0) {
+
+    if (PyArray_IsScalar(sc, Datetime) || PyArray_IsScalar(sc, Timedelta)) {
+        PyArray_DatetimeMetaData *dt_data;
+
+        if (PyArray_IsScalar(sc, Datetime)) {
+            descr = PyArray_DescrNewFromType(NPY_DATETIME);
+        }
+        else {
+            /* Timedelta */
+            descr = PyArray_DescrNewFromType(NPY_TIMEDELTA);
+        }
+        if (descr == NULL) {
+            return NULL;
+        }
+        dt_data = &(((PyArray_DatetimeDTypeMetaData *)descr->c_metadata)->meta);
+        memcpy(dt_data, &((PyDatetimeScalarObject *)sc)->obmeta,
+               sizeof(PyArray_DatetimeMetaData));
+
+        return descr;
+    }
+
+    descr = PyArray_DescrFromTypeObject((PyObject *)Py_TYPE(sc));
+    if (PyDataType_ISUNSIZED(descr)) {
         PyArray_DESCR_REPLACE(descr);
         type_num = descr->type_num;
-        if (type_num == PyArray_STRING) {
+        if (type_num == NPY_STRING) {
             descr->elsize = PyString_GET_SIZE(sc);
         }
-        else if (type_num == PyArray_UNICODE) {
+        else if (type_num == NPY_UNICODE) {
             descr->elsize = PyUnicode_GET_DATA_SIZE(sc);
 #ifndef Py_UNICODE_WIDE
             descr->elsize <<= 1;
 #endif
         }
         else {
-            descr->elsize =
-                ((PyVoidScalarObject *)sc)->ob_size;
-            descr->fields = PyObject_GetAttrString(sc, "fields");
-            if (!descr->fields || !PyDict_Check(descr->fields) ||
-                    (descr->fields == Py_None)) {
-                Py_XDECREF(descr->fields);
-                descr->fields = NULL;
-            }
-            if (descr->fields) {
-                descr->names = PyArray_FieldNames(descr->fields);
+            PyArray_Descr *dtype;
+            dtype = (PyArray_Descr *)PyObject_GetAttrString(sc, "dtype");
+            if (dtype != NULL) {
+                descr->elsize = dtype->elsize;
+                descr->fields = dtype->fields;
+                Py_XINCREF(dtype->fields);
+                descr->names = dtype->names;
+                Py_XINCREF(dtype->names);
+                Py_DECREF(dtype);
             }
             PyErr_Clear();
         }
@@ -575,8 +624,8 @@ PyArray_Scalar(void *data, PyArray_Descr *descr, PyObject *base)
     int swap;
 
     type_num = descr->type_num;
-    if (type_num == PyArray_BOOL) {
-        PyArrayScalar_RETURN_BOOL_FROM_LONG(*(Bool*)data);
+    if (type_num == NPY_BOOL) {
+        PyArrayScalar_RETURN_BOOL_FROM_LONG(*(npy_bool*)data);
     }
     else if (PyDataType_FLAGCHK(descr, NPY_USE_GETITEM)) {
         return descr->f->getitem(data, base);
@@ -585,20 +634,53 @@ PyArray_Scalar(void *data, PyArray_Descr *descr, PyObject *base)
     copyswap = descr->f->copyswap;
     type = descr->typeobj;
     swap = !PyArray_ISNBO(descr->byteorder);
-    if PyTypeNum_ISSTRING(type_num) { /* Eliminate NULL bytes */
-            char *dptr = data;
+    if (PyTypeNum_ISSTRING(type_num)) {
+        /* Eliminate NULL bytes */
+        char *dptr = data;
 
-            dptr += itemsize - 1;
-            while(itemsize && *dptr-- == 0) {
-                itemsize--;
-            }
-            if (type_num == PyArray_UNICODE && itemsize) {
-                /* make sure itemsize is a multiple of 4 */
-                /* so round up to nearest multiple */
-                itemsize = (((itemsize-1) >> 2) + 1) << 2;
-            }
+        dptr += itemsize - 1;
+        while(itemsize && *dptr-- == 0) {
+            itemsize--;
         }
-    if (type->tp_itemsize != 0) { /* String type */
+        if (type_num == NPY_UNICODE && itemsize) {
+            /*
+             * make sure itemsize is a multiple of 4
+             * so round up to nearest multiple
+             */
+            itemsize = (((itemsize - 1) >> 2) + 1) << 2;
+        }
+    }
+#if PY_VERSION_HEX >= 0x03030000
+    if (type_num == NPY_UNICODE) {
+        PyObject *u, *args;
+        int byteorder;
+
+#if NPY_BYTE_ORDER == NPY_LITTLE_ENDIAN
+        byteorder = -1;
+#elif NPY_BYTE_ORDER == NPY_BIG_ENDIAN
+        byteorder = +1;
+#else
+        #error Endianness undefined ?
+#endif
+        if (swap) byteorder *= -1;
+
+        u = PyUnicode_DecodeUTF32(data, itemsize, NULL, &byteorder);
+        if (u == NULL) {
+            return NULL;
+        }
+        args = Py_BuildValue("(O)", u);
+        if (args == NULL) {
+            Py_DECREF(u);
+            return NULL;
+        }
+        obj = type->tp_new(type, args, NULL);
+        Py_DECREF(u);
+        Py_DECREF(args);
+        return obj;
+    }
+#endif
+    if (type->tp_itemsize != 0) {
+        /* String type */
         obj = type->tp_alloc(type, itemsize);
     }
     else {
@@ -607,103 +689,136 @@ PyArray_Scalar(void *data, PyArray_Descr *descr, PyObject *base)
     if (obj == NULL) {
         return NULL;
     }
-    if PyTypeNum_ISFLEXIBLE(type_num) {
-            if (type_num == PyArray_STRING) {
-                destptr = PyString_AS_STRING(obj);
-                ((PyStringObject *)obj)->ob_shash = -1;
-                ((PyStringObject *)obj)->ob_sstate =    \
-                    SSTATE_NOT_INTERNED;
-                memcpy(destptr, data, itemsize);
-                return obj;
-            }
-            else if (type_num == PyArray_UNICODE) {
-                PyUnicodeObject *uni = (PyUnicodeObject*)obj;
-                size_t length = itemsize >> 2;
-#ifndef Py_UNICODE_WIDE
-                char *buffer;
-                int alloc = 0;
-                length *= 2;
-#endif
-                /* Need an extra slot and need to use
-                   Python memory manager */
-                uni->str = NULL;
-                destptr = PyMem_NEW(Py_UNICODE,length+1);
-                if (destptr == NULL) {
-                    Py_DECREF(obj);
-                    return PyErr_NoMemory();
-                }
-                uni->str = (Py_UNICODE *)destptr;
-                uni->str[0] = 0;
-                uni->str[length] = 0;
-                uni->length = length;
-                uni->hash = -1;
-                uni->defenc = NULL;
-#ifdef Py_UNICODE_WIDE
-                memcpy(destptr, data, itemsize);
-                if (swap) {
-                    byte_swap_vector(destptr, length, 4);
-                }
-#else
-                /* need aligned data buffer */
-                if ((swap) || ((((intp)data) % descr->alignment) != 0)) {
-                    buffer = _pya_malloc(itemsize);
-                    if (buffer == NULL) {
-                        return PyErr_NoMemory();
-                    }
-                    alloc = 1;
-                    memcpy(buffer, data, itemsize);
-                    if (swap) {
-                        byte_swap_vector(buffer, itemsize >> 2, 4);
-                    }
-                }
-                else {
-                    buffer = data;
-                }
+    if (PyTypeNum_ISDATETIME(type_num)) {
+        /*
+         * We need to copy the resolution information over to the scalar
+         * Get the void * from the metadata dictionary
+         */
+        PyArray_DatetimeMetaData *dt_data;
 
-                /* Allocated enough for 2-characters per itemsize.
-                   Now convert from the data-buffer
-                */
-                length = PyUCS2Buffer_FromUCS4(uni->str,
-                                               (PyArray_UCS4 *)buffer,
-                                               itemsize >> 2);
-                if (alloc) {
-                    _pya_free(buffer);
-                }
-                /* Resize the unicode result */
-                if (MyPyUnicode_Resize(uni, length) < 0) {
-                    Py_DECREF(obj);
-                    return NULL;
-                }
+        dt_data = &(((PyArray_DatetimeDTypeMetaData *)descr->c_metadata)->meta);
+        memcpy(&(((PyDatetimeScalarObject *)obj)->obmeta), dt_data,
+               sizeof(PyArray_DatetimeMetaData));
+    }
+    if (PyTypeNum_ISFLEXIBLE(type_num)) {
+        if (type_num == NPY_STRING) {
+            destptr = PyString_AS_STRING(obj);
+            ((PyStringObject *)obj)->ob_shash = -1;
+#if !defined(NPY_PY3K)
+            ((PyStringObject *)obj)->ob_sstate = SSTATE_NOT_INTERNED;
 #endif
-                return obj;
+            memcpy(destptr, data, itemsize);
+            return obj;
+        }
+#if PY_VERSION_HEX < 0x03030000
+        else if (type_num == NPY_UNICODE) {
+            /* tp_alloc inherited from Python PyBaseObject_Type */
+            PyUnicodeObject *uni = (PyUnicodeObject*)obj;
+            size_t length = itemsize >> 2;
+            Py_UNICODE *dst;
+#ifndef Py_UNICODE_WIDE
+            char *buffer;
+            Py_UNICODE *tmp;
+            int alloc = 0;
+
+            length *= 2;
+#endif
+            /* Set uni->str so that object can be deallocated on failure */
+            uni->str = NULL;
+            uni->defenc = NULL;
+            uni->hash = -1;
+            dst = PyObject_MALLOC(sizeof(Py_UNICODE) * (length + 1));
+            if (dst == NULL) {
+                Py_DECREF(obj);
+                PyErr_NoMemory();
+                return NULL;
+            }
+#ifdef Py_UNICODE_WIDE
+            memcpy(dst, data, itemsize);
+            if (swap) {
+                byte_swap_vector(dst, length, 4);
+            }
+            uni->str = dst;
+            uni->str[length] = 0;
+            uni->length = length;
+#else
+            /* need aligned data buffer */
+            if ((swap) || ((((npy_intp)data) % descr->alignment) != 0)) {
+                buffer = malloc(itemsize);
+                if (buffer == NULL) {
+                    PyObject_FREE(dst);
+                    Py_DECREF(obj);
+                    PyErr_NoMemory();
+                }
+                alloc = 1;
+                memcpy(buffer, data, itemsize);
+                if (swap) {
+                    byte_swap_vector(buffer, itemsize >> 2, 4);
+                }
             }
             else {
-                PyVoidScalarObject *vobj = (PyVoidScalarObject *)obj;
-                vobj->base = NULL;
-                vobj->descr = descr;
-                Py_INCREF(descr);
-                vobj->obval = NULL;
-                vobj->ob_size = itemsize;
-                vobj->flags = BEHAVED | OWNDATA;
-                swap = 0;
-                if (descr->names) {
-                    if (base) {
-                        Py_INCREF(base);
-                        vobj->base = base;
-                        vobj->flags = PyArray_FLAGS(base);
-                        vobj->flags &= ~OWNDATA;
-                        vobj->obval = data;
-                        return obj;
-                    }
+                buffer = data;
+            }
+
+            /*
+             * Allocated enough for 2-characters per itemsize.
+             * Now convert from the data-buffer
+             */
+            length = PyUCS2Buffer_FromUCS4(dst,
+                    (npy_ucs4 *)buffer, itemsize >> 2);
+            if (alloc) {
+                free(buffer);
+            }
+            /* Resize the unicode result */
+            tmp = PyObject_REALLOC(dst, sizeof(Py_UNICODE)*(length + 1));
+            if (tmp == NULL) {
+                PyObject_FREE(dst);
+                Py_DECREF(obj);
+                return NULL;
+            }
+            uni->str = tmp;
+            uni->str[length] = 0;
+            uni->length = length;
+#endif
+            return obj;
+        }
+#endif /* PY_VERSION_HEX < 0x03030000 */
+        else {
+            PyVoidScalarObject *vobj = (PyVoidScalarObject *)obj;
+            vobj->base = NULL;
+            vobj->descr = descr;
+            Py_INCREF(descr);
+            vobj->obval = NULL;
+            Py_SIZE(vobj) = itemsize;
+            vobj->flags = NPY_ARRAY_CARRAY | NPY_ARRAY_F_CONTIGUOUS | NPY_ARRAY_OWNDATA;
+            swap = 0;
+            if (PyDataType_HASFIELDS(descr)) {
+                if (base) {
+                    Py_INCREF(base);
+                    vobj->base = base;
+                    vobj->flags = PyArray_FLAGS((PyArrayObject *)base);
+                    vobj->flags &= ~NPY_ARRAY_OWNDATA;
+                    vobj->obval = data;
+                    return obj;
                 }
-                destptr = PyDataMem_NEW(itemsize);
-                if (destptr == NULL) {
-                    Py_DECREF(obj);
-                    return PyErr_NoMemory();
-                }
-                vobj->obval = destptr;
+            }
+            destptr = PyDataMem_NEW(itemsize);
+            if (destptr == NULL) {
+                Py_DECREF(obj);
+                return PyErr_NoMemory();
+            }
+            vobj->obval = destptr;
+
+            /*
+             * No base available for copyswp and no swap required.
+             * Copy data directly into dest.
+             */
+            if (base == NULL) {
+                memcpy(destptr, data, itemsize);
+                return obj;
             }
         }
+    }
     else {
         destptr = scalar_value(obj, descr);
     }
@@ -715,9 +830,11 @@ PyArray_Scalar(void *data, PyArray_Descr *descr, PyObject *base)
 /* Return Array Scalar if 0-d array object is encountered */
 
 /*NUMPY_API
-  Return either an array or the appropriate Python object if the array
-  is 0d and matches a Python type.
-*/
+ *
+ * Return either an array or the appropriate Python object if the array
+ * is 0d and matches a Python type.
+ * steals reference to mp
+ */
 NPY_NO_EXPORT PyObject *
 PyArray_Return(PyArrayObject *mp)
 {
@@ -732,9 +849,9 @@ PyArray_Return(PyArrayObject *mp)
     if (!PyArray_Check(mp)) {
         return (PyObject *)mp;
     }
-    if (mp->nd == 0) {
+    if (PyArray_NDIM(mp) == 0) {
         PyObject *ret;
-        ret = PyArray_ToScalar(mp->data, mp);
+        ret = PyArray_ToScalar(PyArray_DATA(mp), mp);
         Py_DECREF(mp);
         return ret;
     }
@@ -742,5 +859,3 @@ PyArray_Return(PyArrayObject *mp)
         return (PyObject *)mp;
     }
 }
-
-
